/*
    cap. Trying to capture from a v4l2 device and display it on X window
    Copyright (C) 2006  Ashish Shukla "Wah Java !!"

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef HAVE_SYS_MMAN_H
#include <sys/mman.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef HAVE_LINUX_VIDEODEV_H
#include <linux/videodev.h>
#endif

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#ifdef HAVE_X11_XLIB_H
#include <X11/Xlib.h>
#endif

#ifdef HAVE_X11_XUTIL_H
#include <X11/Xutil.h>
#endif

#if 0
#ifdef HAVE_X11_EXTENSIONS_XSHM_H
#include <X11/extensions/XShm.h>
#endif
#endif

#ifdef HAVE_X11_EXTENSIONS_XVLIB_H
#include <X11/extensions/Xvlib.h>
#endif

#ifdef HAVE_X11_XUTIL_H
#include <X11/Xutil.h>
#endif

#if 0
#ifdef HAVE_SYS_SHM_H
#include <sys/shm.h>
#endif

#ifdef HAVE_SYS_IPC_H
#include <sys/ipc.h>
#endif
#endif

#ifdef HAVE_SETJMP_H
#include <setjmp.h>
#endif

#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif

#define dump_error() dump_error_at_line(errno, __LINE__)

#ifndef HAVE_GETOPT
#error Create a getopt() clone and add it.
#endif

#ifndef HAVE_BZERO
#ifdef HAVE_MEMSET
#define bzero(ptr,size) memset(ptr, 0, size)
#else
/* non portable across architectures */
#define bzero(ptr,size) __asm__("rep stosb"::"a"(0), "D"(ptr), "c"(size))
#endif
#endif

static void dump_error_at_line(int,int), print_usage(const char*), init_x_atoms(Display*),
	set_text_property(Display*, Window, unsigned char*, unsigned char*), mainloop(Display*);
static int list_inputs(const char*), start_capture(const char*, Display*, int), list_xv_info(Display*),
	x_io_error_handler(Display*);
static int is_device(const char*), is_xv_available(Display*);
static Window create_capture_window(Display*, int, int, const char*);
static const char* get_pixel_format(unsigned int);

#define DEFAULT_VIDEO        "/dev/video"
#define DEFAULT_TITLE        PACKAGE_STRING
#define DEFAULT_X            0
#define DEFAULT_Y            0
#define DEFAULT_BORDER_WIDTH 5
#define DEFAULT_BUFFERS      6
#define DEFAULT_PIX_FORMAT   V4L2_PIX_FMT_YVU420
#define DEFAULT_XV_FORMAT    0x32315659 /* YUV12 Planar compatible with above pix format */

/* T0D0: these are unbound, bound them */
#define DEFAULT_INPUT 0
#define SZ_DEFAULT_INPUT "0"

static Atom aUTF8_STRING;
static char running, error;

#ifdef HAVE_SETJMP_H
static jmp_buf procstate;
#endif

int
main(int argc, char **argv)
{
	int fd_video;
	char* devname = DEFAULT_VIDEO;
	int input = DEFAULT_INPUT;
	char f_listinputs = 0, f_listxvideo = 0;
	Display* display = NULL;
	int ch;

	while((ch = getopt(argc, argv, ":d:i:lLh")) != -1)
	{
		switch(ch)
		{
			case 'd': devname = optarg; break;
			case 'h': print_usage(argv[0]); return 0; break;
			case 'l':
				if(f_listxvideo)
				{
					fputs("-l and -L are exclusive options\n", stderr);
					print_usage(argv[0]);
					return 0;
				}
				f_listinputs = 1; break;
			case 'L':
				if(f_listinputs)
				{
					fputs("-l and -L are exclusive options\n", stderr);
					print_usage(argv[0]);
					return 0;
				}
				f_listxvideo = 1; break;
			case 'i': input = atoi(optarg); break;
			case '?': fputs("Error: invalid option\n", stderr);
				print_usage(argv[0]);
				return 0;
			case ':': fputs("Error: argument needed\n", stderr);
				print_usage(argv[0]);
				return 0;
			default:
				fprintf(stderr, "Error: unexpected behavior on '%c' (%d)\n", ch, ch);
				return -1;
		}
	}

	printf("Device: %s\n", devname);

	if(!is_device(devname))
		return -2;

	if(f_listinputs)
		return list_inputs(devname);

	if(!(display = XOpenDisplay(NULL)))
	{
		fprintf(stderr, "Error: Can't connect to display %s\n", XDisplayName(NULL));
		return -1;
	}

	if(f_listxvideo)
		return list_xv_info(display);

	start_capture(devname, display, input);
}

/* These declarations are in V4L2 specification but not in header ;-) */
#ifndef V4L2_STD_NTSC_443
#define V4L2_STD_NTSC_443 ((v4l2_std_id)0x00004000)
#endif
#ifndef V4L2_STD_NTSC_M_KR
#define V4L2_STD_NTSC_M_KR ((v4l2_std_id)0x00008000)
#endif
#ifndef V4L2_STD_SECAM_LC
#define V4L2_STD_SECAM_LC ((v4l2_std_id)0x00800000)
#endif

int
list_xv_info(Display* display)
{
	Window root;
	XvAdaptorInfo* adaptors;
	XvImageFormatValues* formats;
	XvEncodingInfo* encodings;
	int formats_count, encodings_count, adaptors_count;
	Status status;
	int i, j, k;

	if(!is_xv_available(display))
	{
			XCloseDisplay(display);
			return -2;
	}

	root = DefaultRootWindow(display);

	XvQueryAdaptors(display, root, &adaptors_count, &adaptors);
	printf("No. of adaptors: %d\n=== adaptors ===\n", adaptors_count);
	for(i = 0;  i < adaptors_count; i++)
	{

		printf("Index: %d\nName: %s\nNo. of ports: %ld\nPort Base: %d\nNo. of formats: %ld\n== image formats ==\n",
			   i, adaptors[i].name, adaptors[i].num_ports, adaptors[i].base_id, adaptors[i].num_formats);

		for(j = 0; j < adaptors[i].num_formats; j++)
			printf("Depth: %d\nVisual ID: %ld\n", adaptors[i].formats[j].depth, adaptors[i].formats[j].visual_id);

		puts("== ports ==");
		for(j = 0; j < adaptors[i].num_ports; j++)
		{
			printf("Index: %d\nPort: %d\n", j, j + adaptors[i].base_id);
			puts("= image formats (in 2.2) =");
			formats = XvListImageFormats(display, adaptors[i].base_id + j, &formats_count);

			for(k = 0; k < formats_count; k++)
			{
				printf("Index: %d\nID: 0x%x\nType: %s (%s)\nByte Order: %s endian\nBits Per Pixel: %d\n",
					k, formats[k].id, (formats[k].type == XvRGB ? "RGB":"YUV"),
					(formats[k].format == XvPlanar ? "Planar":"Packed"),
					(formats[k].byte_order == LSBFirst ? "Little":"Big"),
					formats[k].bits_per_pixel);
			}
			XFree(formats);
	
			if((status = XvQueryEncodings(display, j + adaptors[i].base_id, &encodings_count, &encodings)) != Success)
					fprintf(stderr, "Error in XvQueryEncodings: %d\n", status);
			printf("No. of encodings: %ld\n= Encodings =\n", encodings_count);
			for(k = 0; k < encodings_count; k++)
			{
				printf("Index: %d\nEncoding ID: %ld\nName: %s\nDimensions: %dx%d\n",
					k, encodings[k].encoding_id, encodings[k].name, encodings[k].width, encodings[k].height);
			}
			XvFreeEncodingInfo(encodings);
		}

	}
	XvFreeAdaptorInfo(adaptors);
	XCloseDisplay(display);

	return 0;
}

int
start_capture(const char* devname, Display* display, int input)
{
	int fd;
	fd_set set;

	int i;

	Window window;
	GC gc;
	XEvent event;

	XvPortID port;
	XvAdaptorInfo* adaptors = NULL;
	XvImage** xv_buffers = NULL;
	int adaptors_count;

	struct v4l2_input current_input;
	struct v4l2_format current_format;
	struct v4l2_buffer current_buffer;
	struct v4l2_capability current_caps;
	struct v4l2_requestbuffers current_reqbuf;

	if(!is_xv_available(display))
	{
		XCloseDisplay(display);
		return -2;
	}

	fd = open(devname, O_RDWR);

	/* querying for capabilities */
	if(ioctl(fd, VIDIOC_QUERYCAP, &current_caps) == -1)
	{
		if(errno == EINVAL)
		{
			fputs("Error; Invalid video device\n", stderr);
			close(fd);
			return -1;
		}
		else
			dump_error();
	}

	/* does it support capturing ?? */
	if(!(current_caps.capabilities & V4L2_BUF_TYPE_VIDEO_CAPTURE))
	{
		fputs("Error: Not a capture device\n", stderr);
		close(fd);
		return -1;
	}

	/* does it supports streaming I/O ?? */
	if(!(current_caps.capabilities & V4L2_CAP_STREAMING))
	{
		fputs("Error: streaming I/O not supported\n", stderr);
		close(fd);
		return 0;
	}

	/* selecting video input */
	bzero(&current_input, sizeof(struct v4l2_input));
	current_input.index = input;
	if(!ioctl(fd, VIDIOC_ENUMINPUT, &current_input) == -1)
		dump_error();

	printf("Selecting video input: %s\n", current_input.name);

	if(ioctl(fd, VIDIOC_S_INPUT, &current_input.index))
		dump_error();
	/* video input selected */

	/* selecting appropriate format */
	bzero(&current_format, sizeof(struct v4l2_format));
	current_format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if(ioctl(fd, VIDIOC_G_FMT, &current_format) == -1)
		dump_error();

	current_format.fmt.pix.pixelformat = DEFAULT_PIX_FORMAT;

	if(ioctl(fd, VIDIOC_S_FMT, &current_format) == -1)
	{
		fputs("Error: Unsupported video format\n", stderr);
		return -1;
	}
	/* format selected */

	/* request memory buffers for holding video data */
	/* we're doing streaming I/O (memory mapping device buffers) */
	bzero(&current_reqbuf, sizeof(struct v4l2_requestbuffers));
	current_reqbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	current_reqbuf.memory = V4L2_MEMORY_MMAP;
	current_reqbuf.count = DEFAULT_BUFFERS;

	if(ioctl(fd, VIDIOC_REQBUFS, &current_reqbuf) == -1)
	{
		if(errno == EINVAL)
			fputs("Error: memory mapped streaming not supported.\n", stderr);
		else
			dump_error();

		close(fd);
		return -1;
	}

	printf("%d buffer(s) allocated.\n", current_reqbuf.count);

	/* buffers allocated so now start mapping them */

	/* but before mapping do some X stuff */
	init_x_atoms(display);
	window = create_capture_window(display, current_format.fmt.pix.width,
		current_format.fmt.pix.height, DEFAULT_TITLE);
	gc = XCreateGC(display, window, 0, NULL);

	XvQueryAdaptors(display, window, &adaptors_count, &adaptors);
	if(!adaptors_count)
	{
		fputs("Error: no video adaptors found.\n", stderr);
		return -1;
	}
	port = adaptors[0].base_id;
	XvFreeAdaptorInfo(adaptors);

	/* now start mapping */
	xv_buffers = (XvImage**)calloc(current_reqbuf.count, sizeof(XvImage*));

	for(i = 0; i < current_reqbuf.count; i++)
	{
		bzero(&current_buffer, sizeof(struct v4l2_buffer));
		current_buffer.type = current_reqbuf.type;
		current_buffer.memory = V4L2_MEMORY_MMAP;
		current_buffer.index = i;

		/* query for allocated buffer parameters */
		if(ioctl(fd, VIDIOC_QUERYBUF, &current_buffer) == -1)
			dump_error();

		/* just trying to reuse a pointer ;-) */
		if((adaptors = mmap(NULL, current_buffer.length, PROT_READ | PROT_WRITE, MAP_SHARED,
				fd, current_buffer.m.offset)) == MAP_FAILED)
			dump_error();

		printf("Buffer %d mapped at %p\n", i, adaptors);

		/* now allocate an XvImage for Xv */
		if(!(xv_buffers[i] = XvCreateImage(display, port, DEFAULT_XV_FORMAT, (char*)adaptors,
				current_format.fmt.pix.width, current_format.fmt.pix.height)))
		{
			fprintf(stderr, "Error allocating XvImage of %d index\n", i);
			close(fd);
			XDestroyWindow(display, window);
			XCloseDisplay(display);
			return -1;
		}

		/* now queue this buffer */
		if(ioctl(fd, VIDIOC_QBUF, &current_buffer) == -1)
			dump_error();
	}
	adaptors = NULL;

	XSelectInput(display, window, StructureNotifyMask);
	XMapWindow(display, window);
	XSetIOErrorHandler(x_io_error_handler);

	running = 1;
	error = 0;

	/* wait till this window get mapped */
	do
    	XNextEvent(display, &event);
    while(event.type != MapNotify);

#ifdef HAVE_SETJMP
	/* we need this for restoring processor state later in case of cleanup after error */
	if(setjmp(procstate) == 0xBADF00D)
		puts("Doing cleanup");
#endif

	FD_ZERO(&set);
	FD_SET(fd, &set);

	/* start streaming :-) */
	ioctl(fd, VIDIOC_STREAMON, &current_buffer.type);

	current_buffer.index = 0;
	i = 1;
	while(running)
	{
		/* this will block indefinitely, so pass proper timeout argument if timeout is desired */
		switch(select(fd+1, &set, NULL, NULL, NULL))
		{
			case -1: dump_error();  break;
			case 0: fputs("Unexpected timeout\n", stderr); break;
			default:
				if(FD_ISSET(fd, &set))
				{
					current_buffer.type = current_reqbuf.type;
					current_buffer.memory = V4L2_MEMORY_MMAP;
					if(ioctl(fd, VIDIOC_DQBUF, &current_buffer))
					{
						fprintf(stderr, "Error %d: Dequeuing buffer failed due to \"%s\"\n", errno,
							strerror(errno));
						i = 0;
					}
				}
		}
		if(!i)
		{
			i = 1;
			continue;
		}

		/* blit on the screen */
		XvPutImage(display, port, window, gc, (((XvImage**)xv_buffers)[current_buffer.index]), 0, 0,
			xv_buffers[current_buffer.index]->width, xv_buffers[current_buffer.index]->height,
			DEFAULT_X, DEFAULT_Y, xv_buffers[0]->width, xv_buffers[0]->height);

		/* check for pending events */
		if(XPending(display))
    		XNextEvent(display, &event);

		/* now re-queue this buffer */
		if(ioctl(fd, VIDIOC_QBUF, &current_buffer))
			fprintf(stderr, "Error: %d: Queuing buffer failed due to \"%s\"\n", errno,
				strerror(errno));

		i = 1;
	}


	/* stop streaming */
	ioctl(fd, VIDIOC_STREAMOFF, NULL);

	fputs("Unmapping buffers....", stdout);
	for(i = 0; i < current_reqbuf.count; i++)
	{
		munmap(xv_buffers[i]->data, xv_buffers[i]->data_size);
		XFree(xv_buffers[i]);
	}
	puts("done.");
	free(xv_buffers);

	if(!error)
	{
		XFreeGC(display, gc);
		XUnmapWindow(display, window);
		XDestroyWindow(display, window);
		XCloseDisplay(display);
	}

	close(fd);
}

int
x_io_error_handler(Display* display)
{
	printf("I'm here...\n");
	running = 0;
	error = 1;
	longjmp(procstate, 0xBADF00D);
}

void
mainloop(Display* display)
{
	XEvent event;

	while(1)
		XNextEvent(display, &event);
}

int
list_inputs(const char* devname)
{
	int fd;
	struct v4l2_input input;
	struct v4l2_format fmt;
	struct v4l2_capability caps;
	struct v4l2_fmtdesc fmtdesc;

	if((fd = open(devname, O_RDWR))  == -1)
		dump_error();

	bzero(&input, sizeof(struct v4l2_input));
	bzero(&fmt, sizeof(struct v4l2_format));
	bzero(&caps, sizeof(struct v4l2_capability));
	bzero(&fmtdesc, sizeof(struct v4l2_fmtdesc));

	if(ioctl(fd, VIDIOC_QUERYCAP, &caps) == -1)
	{
		if(errno == EINVAL)
		{
			fputs("Error; Invalid video device\n", stderr);
			close(fd);
			return;
		}
		else
			dump_error();
	}

	if(!(caps.capabilities & V4L2_BUF_TYPE_VIDEO_CAPTURE))
	{
		fputs("Error: Not a capture device\n", stderr);
		close(fd);
		return;
	}

	printf("Driver Name: %s\nCard Name: %s\nPlugged in: %s\n I/O Capabilities supported: ",
		   caps.driver, caps.card, caps.bus_info);

	printf("%s%s\n", ((caps.capabilities & V4L2_CAP_READWRITE) ? "read()/write() ":""),
		   ((caps.capabilities & V4L2_CAP_STREAMING) ? "streaming ":""));

	puts("=== inputs ===");

	while(!ioctl(fd, VIDIOC_ENUMINPUT, &input))
	{
		if(errno == EINVAL)
			break;
		printf("%d. %s\n", input.index, input.name);
		printf("Standards: %s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s\n",
			(input.std & V4L2_STD_PAL_B ? "PAL-B ":""),
			(input.std & V4L2_STD_PAL_B1 ? "PAL-B1 ":""),
			(input.std & V4L2_STD_PAL_G ? "PAL-G ":""),
			(input.std & V4L2_STD_PAL_H ? "PAL-H ":""),
			(input.std & V4L2_STD_PAL_I ? "PAL-I ":""),
			(input.std & V4L2_STD_PAL_D ? "PAL-D ":""),
			(input.std & V4L2_STD_PAL_D1 ? "PAL-D1 ":""),
			(input.std & V4L2_STD_PAL_K ? "PAL-K ":""),
			(input.std & V4L2_STD_PAL_M ? "PAL-M ":""),
			(input.std & V4L2_STD_PAL_N ? "PAL-N ":""),
			(input.std & V4L2_STD_PAL_Nc ? "PAL-Nc ":""),
			(input.std & V4L2_STD_PAL_60 ? "PAL-60 ":""),
			(input.std & V4L2_STD_NTSC_M ? "NTSC-M ":""),
			(input.std & V4L2_STD_NTSC_M_JP ? "NTSC-M-JP ":""),
			(input.std & V4L2_STD_NTSC_443 ? "NTSC-443 ":""),
			(input.std & V4L2_STD_NTSC_M_KR ? "NTSC-MR-KR ":""),
			(input.std & V4L2_STD_SECAM_B ? "SECAM-B ":""),
			(input.std & V4L2_STD_SECAM_D ? "SECAM-D ":""),
			(input.std & V4L2_STD_SECAM_G ? "SECAM-G ":""),
			(input.std & V4L2_STD_SECAM_H ? "SECAM-H ":""),
			(input.std & V4L2_STD_SECAM_K ? "SECAM-K ":""),
			(input.std & V4L2_STD_SECAM_K1 ? "SECAM-K1 ":""),
			(input.std & V4L2_STD_SECAM_L ? "SECAM-L ":""),
			(input.std & V4L2_STD_SECAM_LC ? "SECAM-LC ":""),
			(input.std & V4L2_STD_ATSC_8_VSB ? "ATSC/HDTV-8-VSB ":""),
			(input.std & V4L2_STD_ATSC_16_VSB ? "ATSC/HDTV-16-VSB ":""));
		input.index++;
	}

	puts("=== image formats supported ===");
	fmtdesc.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	while(!ioctl(fd, VIDIOC_ENUM_FMT, &fmtdesc))
	{
		if(errno == EINVAL)
		{
			if(fmtdesc.index == 0)
				puts("N/A");
			break;
		}

		printf("Format: %s\n", fmtdesc.description);
		fprintf(stdout, "Pixel format: %s", get_pixel_format(fmtdesc.pixelformat));
	}

	puts("=== current image format ===");
	fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if(!ioctl(fd, VIDIOC_G_FMT, &fmt))
	{
		printf("Dimensions: %dx%d\nSize of image: %d bytes\nBytes per line: %d bytes\nPixel format: %s\n",
			fmt.fmt.pix.width, fmt.fmt.pix.height, fmt.fmt.pix.sizeimage, fmt.fmt.pix.bytesperline,
			get_pixel_format(fmt.fmt.pix.pixelformat));
	}
	else
	{
		if(errno == EINVAL)
			fputs("Error: Unsupported operation\n", stderr);
		else if(errno == EBUSY)
			fputs("Error: I/O in progress\n", stderr);
		else
			dump_error();
	}
	

	close(fd);

	return 0;
}

const char*
get_pixel_format(unsigned int fourcc_code)
{
	switch(fourcc_code)
	{
		case V4L2_PIX_FMT_RGB332: return ("R:3 G:3 B:2"); break;
		case V4L2_PIX_FMT_RGB555: return ("R:5 G:5 B:5"); break;
		case V4L2_PIX_FMT_RGB565: return ("R:5 G:6 B:5"); break;
		case V4L2_PIX_FMT_RGB555X: return ("R:5 G:5 B:5"); break;
		case V4L2_PIX_FMT_RGB565X: return ("R:5 G:6 B:5"); break;
		case V4L2_PIX_FMT_BGR24: return ("B:8 G:8 R:8 (24)"); break;
		case V4L2_PIX_FMT_RGB24: return ("R:8 G:8 B:8 (24)"); break;
		case V4L2_PIX_FMT_BGR32: return ("B:8 G:8 R:8 (32)"); break;
		case V4L2_PIX_FMT_RGB32: return ("R:8 G:8 B:8 (32)"); break;
		case V4L2_PIX_FMT_GREY: return ("YUV (Grey-scale image)"); break;
		case V4L2_PIX_FMT_YUYV: return ("YUV (YUYV 4:2:2)"); break;
		case V4L2_PIX_FMT_UYVY: return ("YUV (UYVY)"); break;
		case V4L2_PIX_FMT_Y41P: return ("YUV (Y41P 4:1:1)"); break;
		case V4L2_PIX_FMT_YVU420: return ("YUV (YVU420 4:2:0)"); break;
		case V4L2_PIX_FMT_YUV420: return ("YUV (YUV420 4:2:0)"); break;
		case V4L2_PIX_FMT_YUV410: return ("YUV (YUV9 4:1:0)"); break;
		case V4L2_PIX_FMT_YVU410: return ("YUV (YVUP 4:1:0)"); break;
		case V4L2_PIX_FMT_YUV422P: return ("YUV (YUV 4:2:2 planar)"); break;
		case V4L2_PIX_FMT_YUV411P: return ("YUV (YUV 4:1:1 planar)"); break;
		case V4L2_PIX_FMT_NV12: return ("YUV (NV12 4:2:0)"); break;
		case V4L2_PIX_FMT_NV21: return ("YUV (NV21 4:2:0)"); break;
		default: return NULL;
	}
}

int
is_xv_available(Display* display)
{
	unsigned int p_version, p_release, p_request_base, p_error_base, p_event_base;
	Status status;

	switch(status = XvQueryExtension(display, &p_version, &p_release, &p_request_base,
		&p_error_base, &p_event_base))
	{
		case Success:
			printf("Xv %d.%d available\n", p_version, p_release);
			return 1;
		case XvBadAlloc:
			fputs("Xv failed due to memory allocation error\n", stderr);
			return 0;
		case XvBadExtension:
			fputs("Xv not available\n", stderr);
			return 0;
		default:
			fprintf(stderr, "Unknown error %d occured while querying for Xv extension.\n", status);
			return 0;
	}
}
void
init_x_atoms(Display* display)
{
	aUTF8_STRING = XInternAtom(display, "UTF8_STRING", False);
}

void 
set_text_property(Display* display, Window window, unsigned char* key, unsigned char* value)
{
	XTextProperty prop;
	Atom aKEY;
	
	aKEY = XInternAtom(display, key, False);
	prop.value = value;
	prop.encoding = aUTF8_STRING;
	prop.format = 8;	
	prop.nitems = strlen(value);
	XSetTextProperty(display, window, &prop, aKEY);
}

Window
create_capture_window(Display* display, int width, int height, const char* title)
{
	Window window;
	Window root;
	int screen;

	root = DefaultRootWindow(display);
	screen = DefaultScreen(display);
	window = XCreateSimpleWindow(display, root, 0, 0, width, height, DEFAULT_BORDER_WIDTH,
								 BlackPixel(display, screen), WhitePixel(display, screen));

	set_text_property(display, window, "WM_NAME", title);
	
	return window;
}

int
is_device(const char* devname)
{
#ifdef HAVE_STAT
	struct stat result;

	if(stat(devname, &result) == -1)
		dump_error();

	if(!S_ISCHR(result.st_mode))
	{
		fprintf(stderr, "Error: \"%s\" is not a character device\n", devname);
		return 0;
	}
#endif
	return 1;
}

void
print_usage(const char* progname)
{
	fprintf(stderr, "%s\nReport bugs to %s\n", PACKAGE_STRING, PACKAGE_BUGREPORT);
	fprintf(stderr, "Usage: %s [options]\n", progname);
	fputs("\t-d <devname> : select video device <devname>, default is \"" DEFAULT_VIDEO "\"\n"
		"\t-l           : lists all available video inputs\n"
		"\t-L           : lists XVideo information\n"
		"\t-v <xv22img> : selects image format with id <xv22img> for XvImages\n"
		"\t-i           : selects video input at <index>, default is \"" SZ_DEFAULT_INPUT "\"\n"
		"\t-h           : displays this message\n\n"
		"NOTE: -l and -L are exclusive options and can't be specified together\n", stderr);
}

void
dump_error_at_line(int err, int line)
{
	fprintf(stderr, "Error %d, Line: %d: %s\n", err, line, strerror(err));
	exit(-1);
}
